from itertools import combinations
import math
import asyncio
import logging

import numpy as np

from .db_actions import write_evaluations_batch
from .io.lammps import LammpsLocalDFTFITCalculator
from .io.lammps_cython import LammpsCythonDFTFITCalculator

logger = logging.getLogger(__name__)


class DFTFITProblemBase:
    def __init__(self, potential, dft_calculations, calculator='lammps', dbm=None, db_write_interval=10, run_id=None, loop=None, **kwargs):
        # Calculator Initialization
        calculator_mapper = {
            'lammps': LammpsLocalDFTFITCalculator,
            'lammps_cython': LammpsCythonDFTFITCalculator,
        }
        self.dft_calculations = dft_calculations
        self.loop = loop or asyncio.get_event_loop()
        self.calculator = calculator_mapper[calculator](structures=[c.structure for c in dft_calculations], potential=potential, **kwargs)
        self.loop.run_until_complete(self.calculator.create())

        # Potential Initialization
        self.potential = potential

        # Database Logging Initialization
        self.dbm = dbm
        self._run_id = run_id
        self.db_write_interval = db_write_interval
        self._evaluation_buffer = []
        if self.dbm and not isinstance(self._run_id, int):
            raise ValueError('cannot write evaluation to database without integer run_id')

    def store_evaluation(self, potential, result):
        if self.dbm:
            self._evaluation_buffer.append([potential, result])
            if len(self._evaluation_buffer) >= self.db_write_interval:
                write_evaluations_batch(self.dbm, self._run_id, self._evaluation_buffer)
                self._evaluation_buffer = []

    def __deepcopy__(self, memo):
        return self # override copy method

    def finalize(self):
        if self._evaluation_buffer: # ensure that all evaluations have been written
            write_evaluations_batch(self.dbm, self._run_id, self._evaluation_buffer)
            self._evaluation_buffer = []

    def __del__(self):
        self.calculator.shutdown()

    def get_bounds(self):
        return tuple(zip(*self.potential.optimization_bounds.tolist()))


class DFTFITSingleProblem(DFTFITProblemBase):
    def __init__(self, w_f, w_s, w_e, **kwargs):
        super().__init__(**kwargs)
        self.weights = {'forces': w_f, 'stress': w_s, 'energy': w_e}

    def get_nobj(self):
        return 1

    def fitness(self, parameters):
        potential = self.potential.copy()
        potential.optimization_parameters = parameters
        md_calculations = self.loop.run_until_complete(self.calculator.submit(potential))
        result = singleobjective_function(self.dft_calculations, md_calculations, self.weights)
        self.store_evaluation(potential, result)
        logger.info(f'evaluation: {result["score"]}')
        return (result['score'],)


class DFTFITMultiProblem(DFTFITProblemBase):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get_nobj(self):
        return 3

    def fitness(self, parameters):
        potential = self.potential.copy()
        potential.optimization_parameters = parameters
        md_calculations = self.loop.run_until_complete(self.calculator.submit(potential))
        result = multiobjective_function(self.dft_calculations, md_calculations)
        self.store_evaluation(potential, result)
        logger.info(f'evaluation: {result["score"]}')
        return result['score']


def multiobjective_function(md_calculations, dft_calculations):
    n_force_sq_error = 0.0
    d_force_sq_error = 0.0
    n_stress_sq_error = 0.0
    d_stress_sq_error = 0.0
    n_energy_sq_error = 0.0
    d_energy_sq_error = 0.0

    for md_calculation, dft_calculation in zip(md_calculations, dft_calculations):
        n_force_sq_error += np.sum((md_calculation.forces - dft_calculation.forces)**2.0)
        d_force_sq_error += np.sum(dft_calculation.forces**2.0)

        n_stress_sq_error += np.sum((md_calculation.stress - dft_calculation.stress)**2.0)
        d_stress_sq_error += np.sum(dft_calculation.stress**2.0)

    if len(md_calculations) > 1:  # cannot calculate energy error if only one set of calculations
        for (md_calc_i, dft_calc_i), (md_calc_j, dft_calc_j) in combinations(zip(md_calculations, dft_calculations), 2):
            n_energy_sq_error += ((md_calc_i.energy - md_calc_j.energy) - (dft_calc_i.energy - dft_calc_j.energy))**2.0
            d_energy_sq_error += (dft_calc_i.energy - dft_calc_j.energy)**2.0
        energy_sq_error = math.sqrt(n_energy_sq_error / d_energy_sq_error)
    else:
        energy_sq_error = 0.0

    force_sq_error = math.sqrt(n_force_sq_error / d_force_sq_error)
    stress_sq_error = math.sqrt(n_stress_sq_error / d_stress_sq_error)

    parts = {'forces': force_sq_error, 'stress': stress_sq_error, 'energy': energy_sq_error}
    score = (force_sq_error, stress_sq_error, energy_sq_error)
    return {'parts': parts, 'score': score}


def singleobjective_function(md_calculations, dft_calculations, weights):
    """ A simple method of scalarizing a multiobjective function:
    https://en.wikipedia.org/wiki/Multi-objective_optimization#Scalarizing

    """
    result = multiobjective_function(md_calculations, dft_calculations)
    result['score'] = (
        weights['forces'] * result['parts']['forces'] + \
        weights['stress'] * result['parts']['stress'] + \
        weights['energy'] * result['parts']['energy']
    )
    result['weights'] = {'forces': weights['forces'], 'stress': weights['stress'], 'energy': weights['energy']}
    return result



from pathlib import Path
from collections import OrderedDict
import asyncio

import numpy as np
from pymatgen.core import Element
from pmg_lammps.output import LammpsRun, LammpsData
from pmg_lammps.inputs import LammpsInput, LammpsScript
from pmg_lammps.calculator.client import LammpsLocalClient

from .base import MDReader, MDCalculator, DFTFITCalculator
from .utils import element_type_to_symbol


class LammpsReader(MDReader):
    def __init__(self, directory, data_filename='initial.data', log_filename='lammps.log', dump_filename='mol.lammpstrj'):
        self.directory = Path(directory)
        self.data_filename = data_filename
        self.log_filename = log_filename
        self.dump_filename = dump_filename
        self._parse()

    def _parse(self):
        dump_path = list(self.directory.glob('**/%s' % self.dump_filename))
        if len(dump_path) > 1:
            raise ValueError('multiple %s files found within directory' % self.dump_filename)
        elif len(dump_path) == 0:
            raise ValueError('could not find %s file within directory' % self.dump_filename)

        log_path = list(self.directory.glob('**/%s' % self.log_filename))
        if len(log_path) > 1:
            raise ValueError('multiple %s files found within directory' % self.log_filename)
        elif len(log_path) == 0:
            raise ValueError('could not find %s file within directory' % self.log_filename)

        data_path = list(self.directory.glob('**/%s' % self.data_filename))
        if len(data_path) > 1:
            raise ValueError('multiple %s files found within directory' % self.data_filename)
        elif len(data_path) == 0:
            raise ValueError('could not find %s file within directory' % self.data_filename)

        data = LammpsData.from_file(str(data_path[0]))
        output = LammpsRun(str(data_path[0]), lammps_log=str(log_path[0]), lammps_dump=str(dump_path[0]))
        # TODO: check units
        self._forces = output.get_forces(-1)
        self._stress = output.get_stress(-1)
        self._energy = output.get_energy(-1)
        self._structure = output.get_structure(-1)


lammps_dftfit_set = OrderedDict([
    ('echo', 'both'),
    ('log', 'lammps.log'),
    ('units', 'metal'),
    ('dimension', 3),
    ('boundary', 'p p p'),
    ('atom_style', 'full'),
    ('read_data', 'initial.data'),
    ('kspace_style', []),
    ('pair_style', []),
    ('pair_coeff', []),
    ('set', []),
    ('dump', '1 all custom 1 mol.lammpstrj id type x y z fx fy fz'),
    ('dump_modify', '1 sort id'),
    ('thermo_modify', 'flush yes'),
    ('thermo_style', 'custom step etotal pxx pyy pzz pxy pxz pyz'),
    ('run', 0),
])


def modify_input_for_potential(lammps_input, potential):
    symbol_indicies = {element_type_to_symbol(s): i for s, i in lammps_input.lammps_data.symbol_indicies.items()}

    # Ensure that even if element missing from structure
    # element is included in lammps script and lammps data
    max_indicie = max(symbol_indicies.values())
    for element in potential.elements:
        if element not in symbol_indicies:
            max_indicie += 1
            elem = Element(element)
            lammps_input.lammps_data.symbol_indicies[elem] = max_indicie
            lammps_input.lammps_data.masses[elem] = elem.atomic_mass
            symbol_indicies[element] = max_indicie

    def charge(potential):
        spec = potential.schema['spec']
        set_commands = []
        element_index_charge = sorted([(symbol_indicies[element], float(charge)) for element, charge in spec.get('charge', {}).items()], key=lambda e: e[0])
        for element_index, charge in element_index_charge:
            set_commands.append('type %d charge %f' % (element_index, charge))
        return ('set', set_commands)

    def kspace_style(potential):
        spec = potential.schema['spec']
        if 'kspace' in spec:
            style = spec['kspace']['type']
            tollerance = spec['kspace']['tollerance']
            return ('kspace_style', '%s %f' % (style, float(tollerance)))
        return ('kspace_style', [])

    def pair_style(potential):
        pair_map = {
            'buckingham': ('buck', '{style} {cutoff}'),
        }

        spec = potential.schema['spec']
        if 'pair' in spec and len(spec['pair']) == 1:
            style, pair_style_format = pair_map[spec['pair'][0]['type']]
            cutoff = spec['pair'][0].get('cutoff', [10.0])[-1]  # angstroms
            if 'kspace'in spec and spec['kspace']['type'] in {'ewald', 'pppm'}:
                style += '/coul/long'
            return ('pair_style', pair_style_format.format(**{'style': style, 'cutoff': float(cutoff)}))
        return ('pair_style', [])

    def pair_coeff(potential):
        spec = potential.schema['spec']
        if 'pair' in spec and len(spec['pair']) == 1:
            if spec['pair'][0]['type'] in ['buckingham']:
                symbols_to_indicies = lambda symbols: [symbol_indicies[s] for s in symbols]
                pair_coeffs = []
                for coeff in spec['pair'][0]['parameters']:
                    # lammps requires that indicies be accending
                    pair_coeffs.append(' '.join(list(map(str, sorted(symbols_to_indicies(coeff['elements'])) + coeff['coefficients']))))
                return ('pair_coeff', pair_coeffs)
        return ('pair_coeff', [])

    if len(potential.schema['spec'].get('pair', [])) > 1:
        raise ValueError('pymatgen-lammps calculator only supports 1 potential (buckingham), lammps-cython calculator is much much much more robust (only here for legacy reasons')

    lammps_input.lammps_script.update([
        charge(potential),
        kspace_style(potential),
        pair_style(potential),
        pair_coeff(potential)
    ])


class LammpsLocalDFTFITCalculator(DFTFITCalculator):
    def __init__(self, structures, potential, command='lammps_ubuntu', num_workers=1):
        self.num_cores = 1
        self.command = command
        self.lammps_local_client = LammpsLocalClient(command=command, num_workers=num_workers)
        self.structures = structures
        self.potential = potential

    async def create(self):
        await self.lammps_local_client.create()

    @staticmethod
    def _convert_to_reader(structure, lammps_result):
            return MDReader(
                forces=np.array(lammps_result['results']['forces']),
                stress=np.array(lammps_result['results']['stress']),
                energy=lammps_result['results']['energy'],
                structure=structure
            )

    async def submit(self, potential, properties=None):
        lammps_set = lammps_dftfit_set
        properties = properties or {'stress', 'energy', 'forces'}

        futures = []
        for structure in self.structures:
            lammps_input = LammpsInput(
                LammpsScript(lammps_set),
                LammpsData.from_structure(structure))
            modify_input_for_potential(lammps_input, potential)
            data_filename = lammps_input.lammps_script.data_filenames[0]
            stdin = str(lammps_input.lammps_script)
            files = {data_filename: str(lammps_input.lammps_data)}
            futures.append(await self.lammps_local_client.submit(
                stdin, files, properties))
        results = await asyncio.gather(*futures)
        return [self._convert_to_reader(s, r) for s, r in zip(self.structures, results)]

    def shutdown(self):
        self.lammps_local_client.shutdown()


class LammpsLocalMDCalculator(MDCalculator):
    def __init__(self, command='lammps_ubuntu', num_workers=1):
        self.num_cores = 1
        self.command = command
        self.lammps_local_client = LammpsLocalClient(command=command, num_workers=num_workers)

    async def create(self):
        await self.lammps_local_client.create()

    async def submit(self, structure, potential, properties=None, lammps_set=None):
        lammps_set = lammps_set or lammps_dftfit_set
        properties = properties or {'stress', 'energy', 'forces'}
        lammps_input = LammpsInput(
            LammpsScript(lammps_set),
            LammpsData.from_structure(structure))
        modify_input_for_potential(lammps_input, potential)
        data_filename = lammps_input.lammps_script.data_filenames[0]
        stdin = str(lammps_input.lammps_script)
        files = {data_filename: str(lammps_input.lammps_data)}
        return await self.lammps_local_client.submit(stdin, files, properties)

    def shutdown(self):
        self.lammps_local_client.shutdown()



import itertools
import collections

import numpy as np
from sklearn import manifold
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import pymatgen as pmg

from .db_actions import filter_evaluations, list_evaluations


def normalize_parameters(optimization_bounds, parameters_array):
    optimization_bounds = np.array(optimization_bounds)
    min_norm_vector = optimization_bounds[:, 0]
    range_norm_vector = optimization_bounds[:, 1] - optimization_bounds[:, 0]
    return (parameters_array - min_norm_vector) / range_norm_vector


def tsne_evaluations(dbm, ax, potential, run_id, limit=1000, condition='best', labels=None):
    cursor = filter_evaluations(dbm, potential, limit, condition, run_id, labels)
    parameters_array = []
    scores = []
    for row in cursor:
        parameters_array.append(row['parameters'])
        scores.append(row['score'])
    parameters_array = np.array(parameters_array)
    scores = np.array(scores)

    with dbm.connection:
        optimization_bounds = dbm.connection.execute(
            'SELECT bounds FROM run WHERE id = ?', (run_id,)).fetchone()['bounds']

    normalized_parameters = normalize_parameters(optimization_bounds, parameters_array)

    tsne = manifold.TSNE(n_components=2, init='pca', random_state=0)
    Y = tsne.fit_transform(normalized_parameters)
    return ax.scatter(Y[:, 0], Y[:, 1], c=scores, cmap=plt.cm.Spectral, alpha=0.2)


def visualize_progress(dbm, run_id, window=100, title=None, filename=None, show=True):
    df = list_evaluations(dbm, run_id)
    title = title or f'Optimization progress for run {run_id}'

    linewidth = 0.5
    fig, axes = plt.subplots(1, 3, sharey=True, sharex=True)
    axes[0].plot(df['score'].rolling(window=window, center=False).mean(), linewidth=linewidth)
    axes[0].set_title('mean')
    axes[0].set_ylabel('normalized score')
    axes[0].set_xlabel('iterations')
    axes[0].ticklabel_format(axis='x', style='sci', scilimits=(-2, 2))
    axes[1].plot(df['score'].rolling(window=window, center=False).median(), linewidth=linewidth)
    axes[1].set_title('median')
    axes[1].set_xlabel('iterations')
    axes[2].plot(df['score'].rolling(window=window, center=False).min(), linewidth=linewidth)
    axes[2].set_title('minimum')
    axes[2].set_xlabel('iterations')
    fig.set_size_inches((15, 5))

    if filename:
        fig.savefig(filename, transparent=True)
    if show:
        plt.show()


def visualize_single_calculation(dft_calculations, md_calculations, plot='forces', show=True, filename=None):
    dft_forces = []
    md_forces = []
    dft_c = []
    md_c = []
    c_ij = [(0, 0), (1, 1), (2, 2), (0, 1), (0, 2), (1, 2)]
    for dft_calc, md_calc in zip(dft_calculations, md_calculations):
        dft_forces.extend(np.linalg.norm(dft_calc.forces, axis=1).tolist())
        md_forces.extend(np.linalg.norm(md_calc.forces, axis=1).tolist())
        dft_c.append([dft_calc.stress[ij] for ij in c_ij])
        md_c.append([md_calc.stress[ij] for ij in c_ij])

    dft_rel_energies = []
    md_rel_energies = []
    for (dft_calc_i, md_calc_i), (dft_calc_j, md_calc_j) in itertools.combinations(zip(dft_calculations, md_calculations), 2):
        dft_rel_energies.append(dft_calc_i.energy - dft_calc_j.energy)
        md_rel_energies.append(md_calc_i.energy - md_calc_j.energy)

    point_alpha = 0.1
    point_size = 3
    def plot_equal_line(axes, x, y):
        axes.plot([np.min(x), np.max(x)], [np.min(x), np.max(x)], '--', color='#E65933', alpha=0.5)

    if plot == 'forces':
        fig, axes = plt.subplots(1, 1)
        axes.scatter(dft_forces, md_forces, alpha=point_alpha, s=point_size)
        axes.set_xlabel(r'DFT force magnitude [$eV/\AA$]')
        axes.set_ylabel(r'MD force magnitude [$eV/\AA$]')
        plot_equal_line(axes, dft_forces, md_forces)
    elif plot == 'energy':
        fig, axes = plt.subplots(1, 1)
        axes.scatter(dft_rel_energies, md_rel_energies, alpha=point_alpha, s=point_size)
        axes.set_xlabel(r'DFT relative energies [$eV$]')
        axes.set_ylabel(r'MD relative energies [$eV$]')
        plot_equal_line(axes, dft_rel_energies, md_rel_energies)
    elif plot == 'stress':
        fig, axes = plt.subplots(2, 3)
        axes[0,0].scatter(dft_c[0], md_c[0])
        axes[0,0].set_xlabel(r'DFT $\sigma_{xx} [bar]$')
        axes[0,0].set_ylabel(r'MD $\sigma_{xx} [bar]$')
        plot_equal_line(axes[0,0], dft_c[0], md_c[0])

        axes[0,1].scatter(dft_c[1], md_c[1])
        axes[0,1].set_xlabel(r'DFT $\sigma_{yy} [bar]$')
        axes[0,1].set_ylabel(r'MD $\sigma_{yy} [bar]$')
        plot_equal_line(axes[0,1], dft_c[1], md_c[1])

        axes[0,2].scatter(dft_c[2], md_c[2])
        axes[0,2].set_xlabel(r'DFT $\sigma_{zz} [bar]$')
        axes[0,2].set_ylabel(r'MD $\sigma_{zz} [bar]$')
        plot_equal_line(axes[0,2], dft_c[2], md_c[2])

        axes[1,0].scatter(dft_c[3], md_c[3])
        axes[1,0].set_xlabel(r'DFT $\sigma_{xy} [bar]$')
        axes[1,0].set_ylabel(r'MD $\sigma_{xy} [bar]$')
        plot_equal_line(axes[1,0], dft_c[3], md_c[3])

        axes[1,1].scatter(dft_c[4], md_c[4])
        axes[1,1].set_xlabel(r'DFT $\sigma_{xz} [bar]$')
        axes[1,1].set_ylabel(r'MD $\sigma_{xz} [bar]$')
        plot_equal_line(axes[1,1], dft_c[4], md_c[4])

        axes[1,2].scatter(dft_c[5], md_c[5])
        axes[1,2].set_xlabel(r'DFT $\sigma_{yz} [bar]$')
        axes[1,2].set_ylabel(r'MD $\sigma_{yz} [bar]$')
        plot_equal_line(axes[1,2], dft_c[5], md_c[5])
        fig.set_size_inches((12, 15))
        plt.tight_layout()

    if filename:
        fig.savefig(filename, transparen=True)
    if show:
        plt.show()
    return fig, axes


def visualize_radial_pair_distribution(calculations, distance=10, filename=None, show=True):
    """visualize the radial pair distributions for each atom in training set"""
    specie_types = set()
    for calculation in calculations:
        specie_types.update(calculation.structure.types_of_specie)

    distances = collections.defaultdict(list)

    for calculation in calculations:
        for original_site, sites in zip(calculation.structure, calculation.structure.get_all_neighbors(10, include_index=True)):
            for (site, distance, index) in sites:
                distances[tuple(sorted([original_site.specie, site.specie]))].append(distance)

    fig, axes = plt.subplots(len(distances), 1, sharex=True, sharey=True)
    for pair, ax in zip(distances, axes.ravel()):
        ax.set_xlabel('distance [angstroms]')
        ax.set_ylabel('count')
        ax.set_title(f'{pair[0].symbol} - {pair[1].symbol}')
        ax.hist(distances[pair], bins=int(distance*10))
    fig.set_size_inches((10, len(distances)*4))
    if filename:
        fig.savefig(filename, transparent=True)
    if show:
        plt.show()
    return fig, axes


def visualize_pair_energies(separations, pair_energies, filename=None, show=True):
    fig, axes = plt.subplots(len(pair_energies), 1, sharex=True, sharey=True)
    for ax, (label, energies) in zip(np.ravel(axes), pair_energies.items()):
        ax.plot(separations, energies)
        ax.axhline(y=0, c='k', linewidth=1, linestyle='dotted')
        ax.set_ylabel('Energy [eV]')
        ax.set_yscale('symlog')
        while len(ax.get_yticks()) > 5: # would be nice if i could just specify 5...
            ax.set_yticks(ax.get_yticks()[::2])
        ax.set_title(label)
    axes[-1].set_xlabel(r'Separation [$\AA$]')
    plt.show()



import pygmo

from .problem import DFTFITSingleProblem, DFTFITMultiProblem


available_algorithms = {
    'pygmo.de': (pygmo.de, 'S'), # S-U
    'pygmo.sade': (pygmo.sade, 'S'), # S-U
    'pygmo.de1220': (pygmo.de1220, 'S'), # S-U
    'pygmo.pso': (pygmo.pso, 'S'), # S-U
    'pygmo.sea': (pygmo.sea, 'S'), # S-U
    'pygmo.sga': (pygmo.sga, 'S'), # S-U
    # 'pygmo.simulated_annealing': simulated_annealing # api does not match others
    'pygmo.bee_colony': (pygmo.bee_colony, 'S'), # S-U
    'pygmo.cmaes': (pygmo.cmaes, 'S'), # S-U
    'pygmo.nsga2': (pygmo.nsga2, 'M'), # M-U
    'pygmo.moead': (pygmo.moead, 'M') # M-U
}


class Optimize:
    def __init__(self, dft_calculations, potential, algorithm='pygmo.de', algorithm_kwargs=None, problem_kwargs=None, dbm=None, run_id=None, db_write_interval=10):
        self.dbm = dbm
        self.algorithm_name = algorithm
        if self.algorithm_name not in available_algorithms:
            raise ValueError(f'algorithm {self.algorithm_name} not available')
        if available_algorithms[self.algorithm_name][1] == 'S':
            internal_problem = DFTFITSingleProblem(potential=potential, dft_calculations=dft_calculations, dbm=dbm, db_write_interval=db_write_interval, run_id=run_id, **problem_kwargs)
        else:
            internal_problem = DFTFITMultiProblem(potential=potential, dft_calculations=dft_calculations, dbm=dbm, db_write_interval=db_write_interval, run_id=run_id, **problem_kwargs)
        self._internal_problem = internal_problem
        self._problem = pygmo.problem(internal_problem)
        self.algorithm_kwargs = algorithm_kwargs or {}

    def population(self, size, seed=None):
        return pygmo.population(self._problem, size, seed=seed)

    def optimize(self, population, steps, seed=None):
        algorithm_constructor = available_algorithms[self.algorithm_name][0]
        self._algorithm = pygmo.algorithm(algorithm_constructor(gen=steps, seed=seed, **self.algorithm_kwargs))
        results = self._algorithm.evolve(population)
        self._internal_problem.finalize()
        return results



""" Handles collection of trial and testing data.

Has a caching layer as to speed up future runs
"""
import json
import shelve
import os

import yaml

from .schema import TrainingSchema
from .io.mattoolkit import MTKReader
from .io.siesta import SiestaReader


class Training:
    def __init__(self, schema, cache_filename=None):
        schema_load, errors = TrainingSchema().load(schema)
        self.schema = schema_load
        self._gather_calculations(cache_filename=cache_filename)

    def _gather_calculations(self, cache_filename=None):
        self._calculations = []
        for calculation in self.schema['spec']:
            if calculation['type'] == 'mattoolkit':
                self._calculations.extend(self.download_mattoolkit_calculations(calculation['selector'], cache_filename=cache_filename))
            elif calculation['type'] == 'Siesta':
                self._calculations.extend(SiestaReader.from_selector(calculation['selector']))

    @property
    def calculations(self):
        return self._calculations

    def __iter__(self):
        return iter(self._calculations)

    def __len__(self):
        return len(self._calculations)

    def download_mattoolkit_calculations(self, selector, cache_filename=None):
        if cache_filename:
            cache_directory, filename = os.path.split(cache_filename)
            os.makedirs(cache_directory, exist_ok=True)
            key = f'mattoolkit.calculation.' + '.'.join(selector['labels'])
            with shelve.open(cache_filename) as cache:
                if key in cache:
                    calc_ids = cache[key]
                else:
                    from mattoolkit.api import CalculationResourceList

                    calculations = CalculationResourceList()
                    calculations.get(params={'labels': selector['labels']})
                    calc_ids = [c.id for c in calculations.items]
                    cache[key] = calc_ids
        else:
            from mattoolkit.api import CalculationResourceList

            calculations = CalculationResourceList()
            calculations.get(params={'labels': selector['labels']})
            calc_ids = [c.id for c in calculations.items]
        return [MTKReader(calc_id, cache_filename=cache_filename) for calc_id in calc_ids]

    @classmethod
    def from_file(cls, filename, format=None, **kwargs):
        if format not in {'json', 'yaml'}:
            if filename.endswith('json'):
                format = 'json'
            elif filename.endswith('yaml') or filename.endswith('yml'):
                format = 'yaml'
            else:
                raise ValueError('unrecognized filetype from filename %s' % filename)

        if format == 'json':
            with open(filename) as f:
                return cls(json.load(f))
        elif format in {'yaml', 'yml'}:
            with open(filename) as f:
                return cls(yaml.load(f), **kwargs)

    def __str__(self):
        return json.dumps(self.schema, sort_keys=True, indent=4)



import asyncio

import pytest

from dftfit.io.lammps_cython import LammpsCythonDFTFITCalculator


@pytest.mark.parametrize('structure_filename, supercell, num_atoms, potential_filename', [
    ('Ne.cif', (2, 2, 2), 32, 'Ne-lennard-jones.yaml'),             # lennard-jones
    ('He.cif', (3, 3, 3), 54, 'He-beck.yaml'),                      # beck
    ('MgO.cif', (2, 2, 2), 64, 'MgO-charge-buck-fitting.yaml'),     # buckingham
    ('MgO.cif', (2, 2, 2), 64, 'MgO-charge-buck-zbl.yaml'),         # zbl
    ('MgO.cif', (2, 2, 2), 64, 'MgO-charge-func.yaml'),             # python-function
    ('LiTaO3.cif', (1, 1, 1), 30, 'LiTaO3-tersoff-2.yaml'),         # tersoff-2
    ('LiTaO3.cif', (1, 1, 1), 30, 'LiTaO3-tersoff-2-charge.yaml'),  # tersoff-2 + charge
    ('3C-SiC.cif', (2, 2, 2), 64, 'SiC-gao-weber.yaml'),            # gao-weber
    ('3C-SiC.cif', (2, 2, 2), 64, 'SiC-tersoff.yaml'),              # tersoff
    ('3C-SiC.cif', (2, 2, 2), 64, 'SiC-vashishta.yaml'),            # vashishta
    ('CdTe.cif', (2, 2, 2), 64, 'CdTe-stillinger-weber.yaml'),      # stillinger-weber
    ('SiO2.cif', (2, 2, 2), 96, 'SiO2-comb.yaml'),                  # comb
    # ('Ti4Cu2O.cif', (1, 1, 1), 112, 'Ti4Cu2O-comb-3.yaml')        # comb-3 (comb-3 freezes...)
])
@pytest.mark.benchmark(group='apply-potential', min_rounds=10)
def test_potential_lammps_cython(
        benchmark, structure, potential,
        structure_filename, supercell, num_atoms, potential_filename):
    s = structure('test_files/structure/%s' % structure_filename) * supercell
    assert len(s) == num_atoms
    p = potential('test_files/potential/%s' % potential_filename)

    calculator = LammpsCythonDFTFITCalculator([s], potential=p)
    loop = asyncio.get_event_loop()
    loop.run_until_complete(calculator.create())

    @benchmark
    def f():
        calculator._apply_potential_files(p)
        calculator.workers[0]._apply_potential(p)



import shelve
import os

import numpy as np

from .base import DFTReader


class MTKReader(DFTReader):
    def __init__(self, calculation_id, cache_filename=None):
        self.calculation_id = calculation_id
        self._load(cache_filename=cache_filename)

    def _download(self, calculation_id):
        from mattoolkit.api.calculation import CalculationResourceItem

        calculation = CalculationResourceItem(calculation_id)
        calculation.get()

        if calculation.format in {'VASP'}:
            results = calculation.results
            return {
                'forces': np.array(results.final_forces),
                'stress': np.array(results.final_stress) * 1e3, # kbar -> bar
                'energy': float(results.final_energy),
                'structure': results.final_structure
            }
        else:
            raise ValueError('unable to use calculation type %s' % calculation.type)

    def _load(self, cache_filename):
        if cache_filename:
            cache_directory, filename = os.path.split(cache_filename)
            os.makedirs(cache_directory, exist_ok=True)
            key = f'mattoolkit.calculation.{self.calculation_id}'
            with shelve.open(cache_filename) as cache:
                if key in cache:
                    results = cache[key]
                else:
                    results = self._download(self.calculation_id)
                    cache[key] = results
        else:
            results = self._download(self.calculation_id)

        self._forces = results['forces']
        self._stress = results['stress']
        self._energy = results['energy']
        self._structure = results['structure']

    @property
    def forces(self):
        return self._forces

    @property
    def stress(self):
        return self._stress

    @property
    def energy(self):
        return self._energy

    @property
    def structure(self):
        return self._structure



from pathlib import Path
from xml.etree import ElementTree
import glob

import numpy as np
import pymatgen as pmg

from .base import DFTReader


class SiestaReader(DFTReader):
    def __init__(self, xml_structure, xml_lattice, xml_energy, xml_stress, xml_forces):
        self._structure = self._parse_structure(xml_structure, xml_lattice)
        self._energy = self._parse_energy(xml_energy)
        self._stress = self._parse_stress(xml_stress)
        self._forces = self._parse_forces(xml_forces)

    @classmethod
    def from_file(cls, directory, output_filename='output.xml', step=-1):
        directory = Path(directory)
        if not directory.is_dir():
            raise ValueError('path %s must exist and be directory' % directory)
        filename = directory / output_filename
        xml_structures, xml_lattices, xml_energies, xml_stresses, xml_forces = cls.xml_parts_from_root(filename)
        return cls(xml_structures[step], xml_lattices[step], xml_energies[step], xml_stresses[step], xml_forces[step])

    @classmethod
    def from_selector(cls, selector):
        data = []
        if 'filename' in selector:
            filename = selector['filename']
            data.extend(cls._from_selector_with_filename(filename, selector))
        elif 'fileglob' in selector:
            for filename in glob.glob(selector['fileglob'], recursive=True):
                data.extend(cls._from_selector_with_filename(filename, selector))
        else:
            raise ValueError('no way for selector to select data need filename or fileglob')
        return data

    @classmethod
    def _from_selector_with_filename(cls, filename, selector):
        filename = Path(filename)
        if not filename.is_file():
            raise ValueError('path %s must exist and be file' % filename)

        xml_structures, xml_lattices, xml_energies, xml_stresses, xml_forces = xml_parts = cls.xml_parts_from_root(filename)
        results = []
        if ('num_samples' in selector) and selector.get('strategy', 'max-separation') == 'max-separation':
            indicies = np.linspace(0, len(xml_structures)-1, selector['num_samples']).astype('int')
            for i in indicies:
                results.append(cls(xml_structures[i], xml_lattices[i], xml_energies[i], xml_stresses[i], xml_forces[i]))
        elif ('num_samples' not in selector) and selector['strategy'] == 'all':
            for xml_parts in zip(xml_structures, xml_lattices, xml_energies, xml_stresses, xml_forces):
                results.append(cls(*xml_parts))
        else:
            raise ValueError('not able to handle selector type yet')
        return results

    @staticmethod
    def xml_parts_from_root(filename):
        namespaces = {'xml': 'http://www.xml-cml.org/schema'}
        root = ElementTree.parse(filename).getroot()

        # throw away last structure (no calculation for last step)
        XML_ATOMARRAY = './/xml:module[@dictRef="MD"]/xml:molecule[1]/xml:atomArray'
        xml_structures = root.findall(XML_ATOMARRAY, namespaces=namespaces)[:-1]

        XML_LATTICE = './/xml:crystal[@title="Lattice Parameters"]'
        xml_lattices = root.findall(XML_LATTICE, namespaces=namespaces)[:-1]

        XML_ENERGIES = './/xml:propertyList[@title="Final KS Energy"]/xml:property[@dictRef="siesta:E_KS"]/xml:scalar'
        xml_energies = root.findall(XML_ENERGIES, namespaces=namespaces)

        XML_STRESS = './/xml:property[@title="Total Stress"]/xml:matrix'
        xml_stresses = root.findall(XML_STRESS, namespaces=namespaces)

        XML_FORCES = './/xml:propertyList[@title="Forces"]/xml:property/xml:matrix'
        xml_forces = root.findall(XML_FORCES, namespaces=namespaces)

        return (xml_structures, xml_lattices, xml_energies, xml_stresses, xml_forces)

    def _parse_structure(self, xml_structure, xml_lattice):
        coordinates = []
        symbols = []
        for atom in xml_structure:
            symbols.append(atom.attrib["elementType"])
            coordinates.append([atom.attrib["x3"], atom.attrib["y3"], atom.attrib["z3"]])
        coordinates = np.array(coordinates).reshape(-1, 3).astype(float)

        namespaces = {'xml': 'http://www.xml-cml.org/schema'}
        XML_LATTICE_LENGTHS = './xml:cellParameter[@parameterType="length"]'
        XML_LATTICE_ANGLES = './xml:cellParameter[@parameterType="angle"]'
        lattice_lengths = [float(_) for _ in xml_lattice.find(XML_LATTICE_LENGTHS, namespaces=namespaces).text.split()]
        lattice_angles = [float(_) for _ in xml_lattice.find(XML_LATTICE_ANGLES, namespaces=namespaces).text.split()]

        lattice = pmg.Lattice.from_parameters(*lattice_lengths, *lattice_angles)
        return pmg.Structure(lattice, symbols, coordinates, coords_are_cartesian=True)

    def _parse_energy(self, xml_energy):
        return float(xml_energy.text)

    def _parse_stress(self, xml_stress):
        eVA32GPa = 160.21766208 # http://greif.geo.berkeley.edu/~driver/conversions.html
        return np.array([float(_) for _ in xml_stress.text.split()]).reshape(3, 3) * eVA32GPa

    def _parse_forces(self, xml_forces):
        return np.array([float(_) for _ in xml_forces.text.split()]).reshape(-1, 3)

    @property
    def forces(self):
        return self._forces

    @property
    def stress(self):
        return self._stress

    @property
    def energy(self):
        return self._energy

    @property
    def structure(self):
        return self._structure



# TODO: update to new interface

import re
import xml.etree.ElementTree as ET
from collections import defaultdict
from math import cos, radians, sqrt

import numpy as np
from ase import Atoms
from ase.geometry import cell_to_cellpar

# Heavily used constants
double_regex = r'[-+]?\d+\.\d+(?:[eE][-+]?\d+)?'
int_regex = '[+-]?\d+'

bohr = 0.529 #Ang/bohr
Ry = 13.60569 #eV/Rydberg
KBar = 0.1 #GPa/KBar




def read_qe_outfile(filename):
    """Reads the output file for a Quantum Espresso run to extract the
    forces on each atom, unit cell, and atom positions.

    Will only output the last scf step. Meaning that if a molecular
    dynamics run was performed only the last step is used.
    """
    with open(filename, "r") as f:
        out_str = f.read()

        alat_regex = r"lattice parameter \(alat\)\s+=\s+({0})".format(double_regex)

        lattice_regex = (
            r"crystal axes: \(cart\. coord\. in units of alat\)\s+"
            r"a\(1\) = \(\s+({0})\s+({0})\s+({0})\s+\)\s+"
            r"a\(2\) = \(\s+({0})\s+({0})\s+({0})\s+\)\s+"
            r"a\(3\) = \(\s+({0})\s+({0})\s+({0})\s+\)\s+"
        ).format(double_regex)

        alat = float(re.search(alat_regex, out_str).group(1))

        basis = np.array([float(_) for _ in re.search(
            lattice_regex, out_str).groups()]) * alat * bohr
        basis.shape = (3, 3)

        force_regex = (
            'atom\s+{1}\s+type\s+{1}\s+force\s+=\s+({0})\s+({0})\s+({0})\s+'
        ).format(double_regex, int_regex)

        force_block_regex = (
            r'Forces acting on atoms \(Ry/au\):\s+'
            r'((?:atom\s+{1}\s+type\s+{1}\s+force\s+=\s+{0}\s+{0}\s+{0}\s+)+)'
        ).format(double_regex, int_regex)

        position_regex = (
            '([A-Z][a-z]?)\s+({0})\s+({0})\s+({0})\s+'
        ).format(double_regex, int_regex)

        position_block_regex = (
            r'ATOMIC_POSITIONS \(angstrom\)\s+'
            r'((?:[A-Z][a-z]?\s+{0}\s+{0}\s+{0}\s+)+)'
        ).format(double_regex)

        stresses_regex = (
            r"total\s+stress\s+\(Ry/bohr\*\*3\)\s+\(kbar\)\s+P=\s+{0}\s+"
            r"({0})\s+({0})\s+({0})\s+{0}\s+{0}\s+{0}\s+"
            r"({0})\s+({0})\s+({0})\s+{0}\s+{0}\s+{0}\s+"
            r"({0})\s+({0})\s+({0})\s+{0}\s+{0}\s+{0}\s+"
        ).format(double_regex)

        total_energy_regex = "!\s+total energy\s+=\s+({0})\s+Ry".format(double_regex)

        steps = []
        for forces_block, positions_block, stresses, total_energy in zip(
                re.findall(force_block_regex, out_str),
                re.findall(position_block_regex, out_str),
                re.findall(stresses_regex, out_str),
                re.findall(total_energy_regex, out_str)):

            positions = []
            symbols = []
            forces = []

            for position, force in zip(
                    re.findall(position_regex, positions_block),
                    re.findall(force_regex, forces_block)):

                symbols.append(position[0])
                positions.append([float(_) for _ in position[1:4]])
                forces.append([float(_) for _ in force])

            forces = np.array(forces)

            stresses = np.array([float(_) for _ in stresses])
            stresses.shape = (3, 3)

            total_energy = float(total_energy)

            steps.append({
                'system': Atoms(symbols=symbols, positions=positions, cell=basis, pbc=True),
                'dft': {
                    'forces': forces * Ry / bohr,
                    'stresses': stresses * Ry / (bohr**3),
                    'total-energy': total_energy * Ry
                }
            })

        if re.search("Molecular Dynamics Calculation", out_str):
            return steps
        else:
            return steps[-1]



import logging

LOG_LEVELS = {'DEBUG', 'INFO', 'WARNING', 'CRITICAL', 'ERROR'}

# Possibly add file handler
def init_logging(level='WARNING'):
    handler = logging.StreamHandler()
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    handler.setFormatter(formatter)
    logging.basicConfig(handlers=[handler], level=level.upper())



import asyncio
import os
import json

import numpy as np

from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.core import Lattice, Structure
from pymatgen.analysis.elasticity import DeformedStructureSet, ElasticTensor, Stress, Strain


from ..io.lammps import LammpsLocalMDCalculator


def load_lammps_set(config_filename):
    from pmg_lammps.sets import MODULE_DIR
    from pmg_lammps.inputs import LammpsScript

    with open(os.path.join(MODULE_DIR, 'sets', config_filename + ".json")) as f:
        return json.load(f, object_pairs_hook=LammpsScript)


class Predict:
    def __init__(self, calculator='lammps', loop=None, **kwargs):
        calculator_mapper = {
            'lammps': LammpsLocalMDCalculator
        }
        self.calculator = calculator_mapper[calculator](**kwargs)
        self.loop = loop or asyncio.get_event_loop()
        self.loop.run_until_complete(self.calculator.create())

    def conventional_structure(self, structure):
        sga = SpacegroupAnalyzer(structure)
        return sga.get_conventional_standard_structure()

    def static(self, structure, potential):
        async def calculate():
            future = await self.calculator.submit(
                structure, potential,
                properties={'forces', 'stress', 'energy'},
                lammps_set=load_lammps_set('static'))
            await future
            return future.result()
        result = self.loop.run_until_complete(calculate())
        return {
            'energy': result['results']['energy'],
            'stress': np.array(result['results']['stress']) * 1e-4,  # convert to GPa
            'forces': np.array(result['results']['forces'])
        }

    def pair(self, specie_a, specie_b, potential, separations):
        max_r = np.max(separations)
        lattice = Lattice.from_parameters(10*max_r, 10*max_r, 10*max_r, 90, 90, 90)
        async def calculate():
            futures = []
            for sep in separations:
                coord_a = (lattice.a*0.5-(sep/2), lattice.b*0.5, lattice.c*0.5)
                coord_b = (lattice.a*0.5+(sep/2), lattice.b*0.5, lattice.c*0.5)
                structure = Structure(
                    lattice,
                    [specie_a, specie_b],
                    [coord_a, coord_b], coords_are_cartesian=True)
                futures.append(await self.calculator.submit(
                    structure, potential,
                    properties={'energy'},
                    lammps_set=load_lammps_set('static')))
            return await asyncio.gather(*futures)
        results = self.loop.run_until_complete(calculate())
        return np.array([r['results']['energy'] for r in results])

    def lattice_constant(self, structure, potential, supercell=(1, 1, 1), etol=1e-6, ftol=1e-6):
        conventional_structure = self.conventional_structure(structure)

        async def calculate():
            relax_lammps_script = load_lammps_set('relax')
            relax_lammps_script['minimize'] = '%f %f 2000 10000' % (etol, ftol)
            future = await self.calculator.submit(
                conventional_structure * supercell, potential,
                properties={'lattice'},
                lammps_set=relax_lammps_script)
            await future
            return future.result()

        result = self.loop.run_until_complete(calculate())
        lattice = Lattice(result['results']['lattice'])
        return conventional_structure.lattice, Lattice(lattice.matrix / np.array(supercell))

    def elastic_constant(self, structure, potential, supercell=(1, 1, 1),
                         nd=0.01, ns=0.05, num_norm=4, num_shear=4,
                         etol=1e-6, ftol=1e-6):
        conventional_structure = self.conventional_structure(structure)
        deformation_set = DeformedStructureSet(conventional_structure * supercell,
                                               nd=nd, ns=ns,
                                               num_norm=num_norm, num_shear=num_shear)
        strains = []
        stresses = []

        async def calculate():
            relax_lammps_script = load_lammps_set('relax')
            relax_lammps_script['fix'] = []
            relax_lammps_script['minimize'] = '%f %f 2000 10000' % (etol, ftol)
            futures = []
            for deformation, deformed_structure in zip(deformation_set.deformations, deformation_set):
                strains.append(Strain.from_deformation(deformation))
                futures.append(await self.calculator.submit(
                    deformed_structure, potential,
                    properties={'stress'},
                    lammps_set=relax_lammps_script))
            for result in await asyncio.gather(*futures):
                stress = Stress(np.array(result['results']['stress']) * 1e-4) # Convert to GPa
                stresses.append(stress)

        self.loop.run_until_complete(calculate())
        return ElasticTensor.from_pseudoinverse(strains, stresses)



import random
import json

import yaml
import os

from .logging import init_logging
from .db import DatabaseManager
from .schema import ConfigurationSchema


class Configuration:
    def __init__(self, schema):
        schema_load, errors = ConfigurationSchema().load(schema)
        self.schema = schema_load

        # Logging
        init_logging(self.schema['spec'].get('logging', 'WARNING'))

        # Name and Labels
        self.run_name = self.schema['metadata'].get('name')
        self.run_labels = self.schema['metadata'].get('labels')

        # Seed
        self.seed = self.schema['spec'].get('seed', random.randint(0, 1_000_000_000))

        # Database
        self.dbm = None
        self.db_write_interval = self.schema['spec'].get('database', {}).get('interval', 10)
        if self.schema['spec'].get('database', {}).get('filename'):
            database_filename = os.path.expanduser(self.schema['spec']['database']['filename'])
            database_directory, filename = os.path.split(database_filename)
            os.makedirs(database_directory, exist_ok=True)
            self.dbm = DatabaseManager(database_filename)

        # Training
        self.training_kwargs = self.schema['spec'].get('training', {
            'cache_filename': '~/.cache/dftfit/cache.db'})

        # Algorithm
        _algorithm_kwargs= self.schema['spec'].get('algorithm', {})
        self.algorithm = _algorithm_kwargs.get('name', 'pygmo.de')
        self.steps = self.schema['spec'].get('steps', 10)
        self.population = self.schema['spec'].get('population', 5)
        self.algorithm_kwargs = {k:v for k,v in _algorithm_kwargs.items() if k != 'name'}

        # Problem
        self.problem_kwargs = self.schema['spec'].get('problem', {})

    @classmethod
    def from_file(cls, filename, format=None):
        if format not in {'json', 'yaml'}:
            if filename.endswith('json'):
                format = 'json'
            elif filename.endswith('yaml') or filename.endswith('yml'):
                format = 'yaml'
            else:
                raise ValueError('unrecognized filetype from filename %s' % filename)

        if format == 'json':
            with open(filename) as f:
                return cls(json.load(f))
        elif format in {'yaml', 'yml'}:
            with open(filename) as f:
                return cls(yaml.load(f))

    def __str__(self):
        return json.dumps(self.schema, sort_keys=True, indent=4)



import os
import argparse

from .utils import is_file_type, is_not_file_type
from ..db import DatabaseManager
from ..db_actions import (
    copy_database_to_database,
    select_potential_from_evaluation,
    list_runs, run_summary
)
from ..visualize import visualize_progress


def add_subcommand_db(subparsers):
    parser = subparsers.add_parser('db', help='database management')
    sub_subparsers = parser.add_subparsers()
    add_subcommand_db_merge(sub_subparsers)
    add_subcommand_db_potential(sub_subparsers)
    add_subcommand_db_progress(sub_subparsers)
    add_subcommand_db_summary(sub_subparsers)


def add_subcommand_db_merge(subparsers):
    parser = subparsers.add_parser('merge', help='merge multiple databases')
    parser.set_defaults(unique=False, func=handle_subcommand_db_merge)
    parser.add_argument('input_databases', nargs='+', type=is_file_type, help='input databases to merge')
    parser.add_argument('-o', '--output-database', help='output database', required=True)
    parser.add_argument('-u', '--unique', dest='unique', action='store_true', help='only insert unique runs')
    parser.add_argument('-f', '--force', dest='force', action='store_true', help='allow to write to existing database file')


def add_subcommand_db_potential(subparsers):
    parser = subparsers.add_parser('potential', help='write evaluation potential to file')
    parser.set_defaults(func=handle_subcommand_db_potential)
    parser.add_argument('database', type=is_file_type, help='database to get potential from')
    parser.add_argument('-e', '--evaluation', type=int, help='use evaluation id')
    parser.add_argument('-o', '--output-filename', type=is_not_file_type, help='output potential filename', required=True)


def add_subcommand_db_progress(subparsers):
    parser = subparsers.add_parser('progress', help='visualize progress of dftfit run')
    parser.set_defaults(func=handle_subcommand_db_progress)
    parser.add_argument('database', type=is_file_type, help='database to get progress from')
    parser.add_argument('--run-id', type=int, help='run id to plot', required=True)
    parser.add_argument('--window', type=int, default=100, help='window to accumulate stats')
    parser.add_argument('--hide', dest='show', action='store_false', help='do not show plot')
    parser.add_argument('-o', '--output-filename', type=is_not_file_type, help='output plot to filename')


def add_subcommand_db_summary(subparsers):
    parser = subparsers.add_parser('summary', help='summary of each run')
    parser.set_defaults(func=handle_subcommand_db_summary)
    parser.add_argument('database', type=is_file_type, help='database to summarize')


def handle_subcommand_db_merge(args):
    if os.path.isfile(args.output_database) and not args.force:
        print(f'path {args.output_database} is an existing file use -f to force writting to existing db')
        exit(1)

    dest_dbm = DatabaseManager(args.output_database)
    for input_database in args.input_databases:
        src_dbm = DatabaseManager(input_database)
        print('merging:', input_database)
        copy_database_to_database(src_dbm, dest_dbm, only_unique=args.unique)


def handle_subcommand_db_potential(args):
    dbm = DatabaseManager(args.database)
    if args.evaluation:
        print(args.evaluation)
        potential = select_potential_from_evaluation(dbm, args.evaluation)
        print(potential)
        potential.write_file(args.output_filename)


def handle_subcommand_db_progress(args):
    dbm = DatabaseManager(args.database)
    visualize_progress(dbm, args.run_id, args.window, filename=args.output_filename, show=args.show)


def handle_subcommand_db_summary(args):
    dbm = DatabaseManager(args.database)
    for run_id in list_runs(dbm):
        r = run_summary(dbm, run_id)
        initial_parameters = ', '.join(['{:4.3g}'.format(_) for _ in r['initial_parameters']])
        final_parameters = ', '.join(['{:4.3g}'.format(_) for _ in r['final_parameters']])
        print(f'''run: {run_id}
        algo: {r['algorithm']:16} steps: {r['steps']:10}
        stats:
              mean:   {r['stats']['mean']:4.3}
              median: {r['stats']['median']:4.3}
              min:    {r['stats']['min']:4.3}
        final:   {final_parameters}
        score:   {r['min_score']}
        ''')



import math
import itertools
import functools
import multiprocessing

import numpy as np

import lammps
from lammps.potential import (
    write_table_pair_potential,
    write_tersoff_potential,
    write_stillinger_weber_potential,
    write_gao_weber_potential,
    write_vashishta_potential,
    write_comb_potential,
    write_comb_3_potential
)

from ..potential import Potential
from .base import DFTFITCalculator, MDReader


class LammpsCythonWorker:
    """A lammps cython worker

    All input and output is fully serializable.
    """
    def __init__(self, structures, elements, potential_schema):
        self.structures = structures
        self.elements = elements
        self.potential = Potential(potential_schema)
        self.lammps_systems = []

    def _initialize_lammps(self, structure):
        lmp = lammps.Lammps(units='metal', style='full', args=[
            '-log', 'none', '-screen', 'none'
        ])
        lmp.system.add_pymatgen_structure(structure, self.elements)
        lmp.thermo.add('my_ke', 'ke', 'all')
        return lmp

    def create(self):
        for structure in self.structures:
            self.lammps_systems.append(self._initialize_lammps(structure))

    def _apply_potential(self, potential):
        lammps_commands = write_potential(potential, elements=self.elements, unique_id=1)
        for command in lammps_commands:
            for lmp in self.lammps_systems:
                lmp.command(command)

    def worker_multiprocessing_loop(self, pipe):
        while True:
            message = pipe.recv()
            if message == 'quit':
                break
            results = self.compute(message)
            pipe.send(results)
        pipe.close()

    def compute(self, parameters):
        self.potential.optimization_parameters = parameters
        self._apply_potential(self.potential)
        results = []
        for lmp in self.lammps_systems:
            lmp.run(0)
            S = lmp.thermo.computes['thermo_press'].vector
            results.append({
                'forces': lmp.system.forces.copy(),
                'energy': lmp.thermo.computes['thermo_pe'].scalar + lmp.thermo.computes['my_ke'].scalar,
                'stress': np.array([
                    [S[0], S[3], S[5]],
                    [S[3], S[1], S[4]],
                    [S[5], S[4], S[2]]
                ])
            })
        return results


class LammpsCythonDFTFITCalculator(DFTFITCalculator):
    """This is not a general purpose lammps calculator. Only for dftfit
    evaluations. For now there are not plans to generalize it.
    """
    def __init__(self, structures, potential, num_workers=1):
        self.structures = structures

        # ensure element indexes are the same between all lammps calculations
        self.elements = set()
        for structure in self.structures:
            self.elements = self.elements | set(structure.species)
        self.elements = list(self.elements)

        self.workers = []
        potential_schema = potential.as_dict()
        if num_workers == 1:
            self.workers.append(LammpsCythonWorker(structures, self.elements, potential_schema))
        else:
            def create_worker(structures, elements, potential_schema, pipe):
                worker = LammpsCythonWorker(structures, elements, potential_schema)
                worker.create()
                worker.worker_multiprocessing_loop(pipe)

            self.workers = []
            structure_index = 0
            rem = len(structures) % num_workers
            n = math.floor(len(structures) / num_workers)
            for i in range(num_workers):
                p_conn, c_conn = multiprocessing.Pipe()
                # hand out remaining to first rem < i
                if num_workers - rem >= i:
                    subset_structures = structures[structure_index: structure_index+n+1]
                    structure_index += n + 1
                else:
                    subset_structures = structures[structure_index: structure_index+n]
                    structure_index += n
                p = multiprocessing.Process(target=create_worker, args=(subset_structures, self.elements, potential_schema, c_conn))
                p.start()
                self.workers.append((p, p_conn))

    async def create(self):
        # otherwise seperate process calls this method
        if len(self.workers) == 1:
            self.workers[0].create()

    def _apply_potential_files(self, potential):
        lammps_files = write_potential_files(potential, elements=self.elements, unique_id=1)
        for filename, content in lammps_files.items():
            with open(filename, 'w') as f:
                f.write(content)

    async def submit(self, potential, properties=None):
        properties = properties or {'stress', 'energy', 'forces'}
        parameters = potential.optimization_parameters
        self._apply_potential_files(potential)

        results = []
        if len(self.workers) == 1:
            results = self.workers[0].compute(parameters)
        else:
            # send potential to each worker
            for p, p_conn in self.workers:
                p_conn.send(parameters)

            # recv calculation results from each worker
            for p, p_conn in self.workers:
                results.extend(p_conn.recv())

        md_readers = []
        for structure, result in zip(self.structures, results):
            md_readers.append(MDReader(energy=result['energy'], forces=result['forces'], stress=result['stress'], structure=structure))
        return md_readers

    def shutdown(self):
        # nothing is needed if not using multiprocessing module
        if len(self.workers) > 1:
            for p, p_conn in self.workers:
                p_conn.send('quit')
                p.join()


def tersoff_2_to_tersoff(element_parameters, mixing_parameters):
    def mixing_params_from_singles(e1, e2, mixing_value):
        p1 = [float(_) for _ in element_parameters[e1]]
        p2 = [float(_) for _ in element_parameters[e2]]
        mixing = float(mixing_value)
        return [
            3.0,                                # m
            1.0,                                # gamma
            0.0,                                # lambda3
            p1[0],                              # c
            p1[1],                              # d
            p1[2],                              # costheta0
            p1[3],                              # n
            p1[4],                              # beta
            (p1[5] + p2[5]) / 2,                # lambda2
            mixing * math.sqrt(p1[6] * p2[6]),  # B
            math.sqrt(p1[7] * p2[7]),           # R
            math.sqrt(p1[8] * p2[8]),           # D
            (p1[9] + p2[9]) / 2,                # lambda1
            math.sqrt(p1[10] * p2[10]),         # A
        ]

    parameters = {}
    for e1, e2, e3 in itertools.product(element_parameters, repeat=3):
        if e1 == e2:
            mixing_value = 1.0
        else:
            sorted_e1_e2 = tuple(sorted([e1, e2]))
            mixing_value = mixing_parameters.get(sorted_e1_e2)
            if mixing_value is None:
                continue
        parameters[(e1, e2, e3)] = mixing_params_from_singles(e1, e2, mixing_value)
    return parameters


LAMMPS_POTENTIAL_NAME_MAPPING = {
    'lennard-jones': 'lj/cut',
    'beck': 'beck',
    'zbl': 'zbl',
    'buckingham': 'buck',
    'tersoff-2': 'tersoff',
    'tersoff': 'tersoff',
    'stillinger-weber': 'sw',
    'gao-weber': 'gw',
    'vashishta': 'vashishta',
    'comb': 'comb',
    'comb-3': 'comb3',
    'python-function': 'table'
}


def write_potential_files(potential, elements, unique_id=1):
    """Generate lammps files required by specified potential

    Parameters
    ----------
    potential: dftfit.potential.Potential
        schema representation of potential
    elements: list
        list specifying the index of each element
    unique_id: int
        an id that can be used for files to guarentee uniqueness
    """
    spec = potential.schema['spec']
    lammps_files = {}
    for i, pair_potential in enumerate(spec.get('pair', [])):
        potential_lammps_name = LAMMPS_POTENTIAL_NAME_MAPPING.get(pair_potential['type'])
        if pair_potential['type'] == 'tersoff-2':
            element_parameters = {}
            mixing_parameters = {}
            for parameter in pair_potential['parameters']:
                if len(parameter['elements']) == 1:
                    element_parameters[parameter['elements'][0]] = parameter['coefficients']
                elif len(parameter['elements']) == 2:
                    mixing_parameters[tuple(sorted(parameter['elements']))] = parameter['coefficients'][0]
            parameters = tersoff_2_to_tersoff(element_parameters, mixing_parameters)
        elif pair_potential['type'] in {'tersoff-2', 'tersoff', 'stillinger-weber', 'gao-weber', 'vashishta', 'comb', 'comb-3', 'python-function'}:
            parameters = {}
            for parameter in pair_potential['parameters']:
                parameters[tuple(parameter['elements'])] = [float(_) for _ in parameter['coefficients']]

        filename = '/tmp/lammps.%d.%d.%s' % (i, unique_id, potential_lammps_name)
        if pair_potential['type'] in {'tersoff-2', 'tersoff'}:
            lammps_files[filename] = write_tersoff_potential(parameters)
        elif pair_potential['type'] == 'stillinger-weber':
            lammps_files[filename] = write_stillinger_weber_potential(parameters)
        elif pair_potential['type'] == 'gao-weber':
            lammps_files[filename] = write_gao_weber_potential(parameters)
        elif pair_potential['type'] == 'vashishta':
            lammps_files[filename] = write_vashishta_potential(parameters)
        elif pair_potential['type'] == 'comb':
            lammps_files[filename] = write_comb_potential(parameters)
        elif pair_potential['type'] == 'comb-3':
            lammps_files[filename] = write_comb_3_potential(parameters)
        elif pair_potential['type'] == 'python-function':
            cutoff = [float(_) for _ in pair_potential.get('cutoff', [1.0, 10.0])]
            samples = int(pair_potential.get('samples', 1000))

            def get_function(func_str):
                d = {}
                exec(func_str, d)
                return d['potential']

            potential_func = get_function(pair_potential['function'])
            for (e1, e2), parameters in parameters.items():
                float_parameters = [float(_) for _ in parameters]
                f_r = functools.partial(potential_func, *float_parameters)
                filename = '/tmp/lammps.%s.%s.%d.%d.%s' % (e1, e2, i, unique_id, potential_lammps_name)
                lammps_files[filename] = write_table_pair_potential(f_r, samples=samples, bounds=cutoff)
    return lammps_files


def write_potential(potential, elements, unique_id=1):
    """Generate lammps commands required by specified potential

    Parameters
    ----------
    potential: dftfit.potential.Potential
        schema representation of potential
    elements: list
        list specifying the index of each element
    unique_id: int
        an id that can be used for files to guarentee uniqueness

    Supported Potentials:

    - two-body
      - lennard-jones
      - buckingham
    - three-body
      - tersoff-2, tersoff
      - stillinger-weber
      - gao-weber
    - n-body
      - coloumb charge (long + short range)
    """
    spec = potential.schema['spec']
    element_map = {e.symbol: i for i, e in enumerate(elements, start=1)}

    # collect potentials in spec
    potentials = []
    lammps_commands = []
    if ('charge' in spec) and ('kspace' in spec):
        lammps_commands.append('kspace_style %s %f' % (spec['kspace']['type'], spec['kspace']['tollerance']))
        for element, charge in spec['charge'].items():
            lammps_commands.append('set type %d charge %f' % (element_map[element], float(charge)))
        potentials.append(({
            'pair_style': 'coul/long %f' % 10.0,
            'pair_coeff': [('* *', 'coul/long', '')]
        }))

    for i, pair_potential in enumerate(spec.get('pair', [])):
        potential_lammps_name = LAMMPS_POTENTIAL_NAME_MAPPING.get(pair_potential['type'])
        if pair_potential['type'] in {'lennard-jones', 'beck', 'buckingham', 'zbl'}:
            pair_coeffs = []
            for parameter in pair_potential['parameters']:
                e1, e2 = parameter['elements']
                ij = ' '.join([str(_) for _ in sorted([element_map[e1], element_map[e2]])])
                coefficients_str = ' '.join([str(float(coeff)) for coeff in parameter['coefficients']])
                pair_coeffs.append((ij, potential_lammps_name, coefficients_str))

            if pair_potential['type'] == 'zbl':
                cutoff = pair_potential.get('cutoff', [3.0, 4.0])
                pair_style = '%s %f %f' % (potential_lammps_name, cutoff[0], cutoff[1])
            else:
                pair_style = '%s %f' % (potential_lammps_name, pair_potential.get('cutoff', [10.0])[-1])
            potentials.append({
                'pair_style': pair_style,
                'pair_coeff': pair_coeffs
            })
        elif pair_potential['type'] == 'python-function':
            pair_coeffs = []
            for parameter in pair_potential['parameters']:
                e1, e2 = parameter['elements']
                ij = ' '.join([str(_) for _ in sorted([element_map[e1], element_map[e2]])])
                filename = '/tmp/lammps.%s.%s.%d.%d.%s' % (e1, e2, i, unique_id, potential_lammps_name)
                pair_coeffs.append((ij, potential_lammps_name, '%s PAIR' % filename))

            samples = pair_potential.get('samples', 1000)
            potentials.append({
                'pair_style': 'table linear %d' % samples,
                'pair_coeff': pair_coeffs
            })
        elif pair_potential['type'] in {'tersoff-2', 'tersoff', 'stillinger-weber', 'gao-weber', 'vashishta', 'comb', 'comb-3'}:
            filename = '/tmp/lammps.%d.%d.%s' % (i, unique_id, potential_lammps_name)
            if pair_potential['type'] == 'comb-3':
                pair_style = '%s polar_off' % (potential_lammps_name)
            else:
                pair_style = potential_lammps_name
            potentials.append({
                'pair_style': pair_style,
                'pair_coeff': [('* *', potential_lammps_name, '%s %s' % (
                    filename, ' '.join(str(e) for e in elements)))],
            })
        else:
            raise ValueError('pair potential %s not implemented yet!' % (pair_potential['type']))

    if len(potentials) == 1:  # no need for hybrid/overlay
        potential = potentials[0]
        lammps_commands.append('pair_style %s' % potential['pair_style'])
        for pair_coeff in potential['pair_coeff']:
            lammps_commands.append('pair_coeff ' + pair_coeff[0] + ' ' + pair_coeff[2])
    elif len(potentials) > 1:  # use hybrid/overlay to join all potentials
        lammps_commands.append('pair_style hybrid/overlay ' + ' '.join(potential['pair_style'] for potential in potentials))
        for potential in potentials:
            for pair_coeff in potential.get('pair_coeff', []):
                lammps_commands.append('pair_coeff ' + ' '.join(pair_coeff))
    return lammps_commands
